#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sched.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <idsa_internal.h>

struct log_action_state {
  int a_fd;
  char a_name[IDSA_M_MEDIUM];	/* name of target */

  off_t a_rotate;		/* rotate: when do we trigger */
  off_t a_have;			/* rotate: how much do we have currently */
  int a_rd;			/* rotate: alternate descriptor */

  int a_sync;			/* synchronous writes */

  pid_t a_pid;			/* pid of pipe */

  IDSA_PRINT_HANDLE *a_handle;
};

/****************************************************************************/

static void idsa_module_log_file(IDSA_RULE_CHAIN * c, struct log_action_state *a);
static void idsa_module_log_pipe(IDSA_RULE_CHAIN * c, struct log_action_state *a);

static int idsa_module_log_open(IDSA_RULE_CHAIN * c, char *name, int flags);
static int idsa_module_log_strexec(char *s);


/****************************************************************************/

void *idsa_module_log_astart(IDSA_MEX_STATE * m, IDSA_RULE_CHAIN * c, void *g)
{
  struct log_action_state *result;
  IDSA_MEX_TOKEN *type, *token;

  result = malloc(sizeof(struct log_action_state));
  if (result == NULL) {
    idsa_chain_error_malloc(c, sizeof(struct log_action_state));
    return NULL;
  }

  result->a_fd = (-1);
  result->a_rd = (-1);

  result->a_rotate = 0;
  result->a_have = 0;

  result->a_pid = 0;
  result->a_handle = NULL;

  result->a_sync = 0;

  type = idsa_mex_get(m);

  if (type == NULL) {
    idsa_chain_error_mex(c, m);
    return result;
  }

  token = idsa_mex_get(m);
  if (token == NULL) {
    idsa_chain_error_mex(c, m);
    return result;
  }

  if (token->t_buf[0] != '/') {
    idsa_chain_error_usage(c, "log destination <%s> has to be an absolute path", token->t_buf);
    return result;
  }
  strncpy(result->a_name, token->t_buf, IDSA_M_MEDIUM - 1);
  result->a_name[IDSA_M_MEDIUM - 1] = '\0';

  /* collect all the options */
  token = idsa_mex_get(m);
  while (token != NULL) {
    if (token->t_id != IDSA_PARSE_COMMA) {	/* no more options, go back */
      idsa_mex_unget(m, token);
      token = NULL;
    } else {
      token = idsa_mex_get(m);
      if (token) {

	if (strcmp(token->t_buf, "rotate") == 0) {
	  token = idsa_mex_get(m);
	  if (token) {
	    result->a_rotate = atoi(token->t_buf);
	    if (result->a_rotate == 0) {
	      idsa_chain_error_usage(c, "expected a nonzero rotation value instead of <%s>", token->t_buf);
	      return result;
	    }
	  } else {
	    idsa_chain_error_mex(c, m);
	    return result;
	  }
	} else if (strcmp(token->t_buf, "custom") == 0) {
	  if (result->a_handle) {
	    idsa_print_free(result->a_handle);
	    result->a_handle = NULL;
	  }
	  token = idsa_mex_get(m);
	  if (token) {
	    result->a_handle = idsa_print_parse(token->t_buf);
	    if (result->a_handle == NULL) {
	      idsa_chain_error_usage(c, "unable to process custom output format <%s>", token->t_buf);
	      return result;
	    }
	  } else {
	    idsa_chain_error_mex(c, m);
	    return result;
	  }
	} else if (strcmp(token->t_buf, "format") == 0) {
	  if (result->a_handle) {
	    idsa_print_free(result->a_handle);
	    result->a_handle = NULL;
	  }
	  token = idsa_mex_get(m);
	  if (token) {
	    result->a_handle = idsa_print_format(token->t_buf);
	    if (result->a_handle == NULL) {
	      idsa_chain_error_usage(c, "unable to process predefined output format <%s>", token->t_buf);
	      return result;
	    }
	  } else {
	    idsa_chain_error_mex(c, m);
	    return result;
	  }
	} else if (strcmp(token->t_buf, "sync") == 0) {
	  result->a_sync = 1;
	} else {
	  idsa_chain_error_usage(c, "unknown log option <%s>", token->t_buf);
	  return result;
	}
	token = idsa_mex_get(m);

      } else {
	idsa_chain_error_mex(c, m);
	return result;
      }
    }
  }

  if (strcmp(type->t_buf, "file") == 0) {
    idsa_module_log_file(c, result);
  } else if (strcmp(type->t_buf, "pipe") == 0) {
    idsa_module_log_pipe(c, result);
  } else {
    idsa_chain_error_usage(c, "unknown log type <%s>", type->t_buf);
    return result;
  }

  return result;
}

int idsa_module_log_acache(IDSA_MEX_STATE * m, IDSA_RULE_CHAIN * c, void *g, void *a)
{
  IDSA_MEX_TOKEN *type, *target, *comma;
  int result;
  struct log_action_state *action;

  action = (struct log_action_state *) (a);

  type = idsa_mex_get(m);
  if (type == NULL) {
    idsa_chain_error_mex(c, m);
    return 1;
  }

  /* bit icky, a file is greater than a pipe (nonzero pid) */
  if (strcmp(type->t_buf, "file") == 0) {
    if (action->a_pid) {
      return 1;
    }
  } else if (strcmp(type->t_buf, "pipe") == 0) {
    if (!action->a_pid) {
      return -1;
    }
  }

  target = idsa_mex_get(m);
  if (target == NULL) {
    idsa_chain_error_mex(c, m);
    return 1;
  }

  result = strcmp(target->t_buf, action->a_name);
  if (result == 0) {
    comma = idsa_mex_peek(m);
    if (comma && (comma->t_id == IDSA_PARSE_COMMA)) {
      idsa_chain_error_usage(c, "options for \"log %s %s\" can only be specified at the first mention", type->t_buf, target->t_buf);
    }
  }

  return result;
}

#define BUFFER (8*IDSA_M_MESSAGE)

int idsa_module_log_ado(IDSA_RULE_CHAIN * c, void *g, void *a, IDSA_EVENT * q, IDSA_EVENT * p)
{
  struct log_action_state *action;
  char buffer[BUFFER];
  int wr, sw;

  action = (struct log_action_state *) (a);

  sw = idsa_print_do(q, action->a_handle, buffer, BUFFER);
  if (sw <= 0) {
    idsa_chain_error_internal(c, "nothing to write to <%s>", action->a_name);
    return 1;
  }

  wr = write(action->a_fd, buffer, sw);
  if (wr != sw) {
    idsa_chain_error_system(c, errno, "write to <%s> failed", action->a_name);
    return 1;
  }

  if (action->a_rotate) {
    action->a_have += sw;
    if (action->a_have >= action->a_rotate) {
      int td;

      td = action->a_fd;
      action->a_fd = action->a_rd;
      action->a_rd = td;
      action->a_have = 0;

      if (ftruncate(action->a_fd, 0)) {
	idsa_chain_error_system(c, errno, "truncate of <%s> failed", action->a_name);
	return 1;
      }
    }
  }

  return 0;
}

void idsa_module_log_astop(IDSA_RULE_CHAIN * c, void *g, void *a)
{
  struct log_action_state *action;

  action = (struct log_action_state *) (a);

  if (action) {
    if (action->a_fd != (-1)) {
      close(action->a_fd);
    }
    if (action->a_rd != (-1)) {
      close(action->a_rd);
    }
    if (action->a_handle != NULL) {
      idsa_print_free(action->a_handle);
      action->a_handle = NULL;
    }

    free(action);
  }
}

/****************************************************************************/

static void idsa_module_log_file(IDSA_RULE_CHAIN * c, struct log_action_state *a)
{
  int flags;

  if (a->a_handle == NULL) {
    a->a_handle = idsa_print_format("native");
    if (a->a_handle == NULL) {
      idsa_chain_error_internal(c, "unable to intialize output format for %s", a->a_name);
    }
  }

  flags = O_APPEND | O_CREAT | O_WRONLY;
#ifdef O_SYNC
  if (a->a_sync)
    flags |= O_SYNC;
#else
#warning O_SYNC not available, synchronous logging option will not be honoured
#endif

  if (a->a_rotate) {		/* should rotate */
    char buf[IDSA_M_MEDIUM + 3];
    struct stat fst, rst;
    int td;

    snprintf(buf, IDSA_M_MEDIUM + 2, "%s-1", a->a_name);
    buf[IDSA_M_MEDIUM + 2] = '\0';
    a->a_fd = idsa_module_log_open(c, buf, flags);
    if (fstat(a->a_fd, &fst)) {
      idsa_chain_error_system(c, errno, "unable to stat <%s>", buf);
    }

    snprintf(buf, IDSA_M_MEDIUM + 2, "%s-2", a->a_name);
    buf[IDSA_M_MEDIUM + 2] = '\0';
    a->a_rd = idsa_module_log_open(c, buf, flags);
    if (fstat(a->a_rd, &rst)) {
      idsa_chain_error_system(c, errno, "unable to stat <%s>", buf);
    }

    /* work with newest file */
    if (rst.st_mtime > fst.st_mtime) {
      td = a->a_rd;
      a->a_rd = a->a_fd;
      a->a_fd = td;

      a->a_have = rst.st_size;
    } else {
      a->a_have = fst.st_size;
    }

  } else {
    a->a_fd = idsa_module_log_open(c, a->a_name, flags);;
  }

}

static int idsa_module_log_open(IDSA_RULE_CHAIN * c, char *name, int flags)
{
  int result;
  mode_t mode;

  mode = S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH;

  result = open(name, flags, mode);
  if (result == (-1)) {
    idsa_chain_error_system(c, errno, "unable to open file <%s>", name);
  } else {
    fcntl(result, F_SETFD, FD_CLOEXEC);
  }

  return result;
}

static void idsa_module_log_pipe(IDSA_RULE_CHAIN * c, struct log_action_state *a)
{
  int p[2];
  int status;
  int flags;

  if (a->a_handle == NULL) {
    a->a_handle = idsa_print_format("internal");
    if (a->a_handle == NULL) {
      idsa_chain_error_internal(c, "unable to intialize output format for %s", a->a_name);
    }
  }

  if (a->a_rotate) {
    idsa_chain_error_usage(c, "pipes do not allow rotation");
    return;
  }

  if (pipe(p)) {
    idsa_chain_error_system(c, errno, "unable to create pipe for <%s>", a->a_name);
    return;
  }

  a->a_pid = fork();
  switch (a->a_pid) {
  case -1:
    idsa_chain_error_system(c, errno, "unable to fork <%s>", a->a_name);
    close(p[0]);
    close(p[1]);
    break;
  case 0:			/* in child */

    close(STDOUT_FILENO);
    close(STDERR_FILENO);
    close(p[1]);

    if (p[0] != STDIN_FILENO) {
      if (dup2(p[0], STDIN_FILENO) != STDIN_FILENO) {	/* failure */
	/* WARNING: abuse of return code */
	exit(errno);
      } else {
	close(p[0]);
      }
    }

    idsa_module_log_strexec(a->a_name);
    /* WARNING: abuse return code again */
    exit(errno);

    break;
  default:			/* in parent */
    close(p[0]);
    a->a_fd = p[1];

    /* yield to child */
    sched_yield();

    if (waitpid(a->a_pid, &status, WNOHANG)) {	/* found a child or error */
      if (WIFEXITED(status)) {
	idsa_chain_error_system(c, WEXITSTATUS(status), "unable to start child process <%s>", a->a_name);
      } else if (WIFSIGNALED(status)) {
	idsa_chain_error_internal(c, "child process <%s> killed by signal %d", a->a_name, WTERMSIG(status));
      }
    }

    if (a->a_sync == 0) {
      flags = fcntl(a->a_fd, F_GETFL, 0);
      if (flags != (-1)) {
	fcntl(a->a_fd, F_SETFL, O_NONBLOCK | flags);
      } else {
	idsa_chain_error_system(c, errno, "unable to make pipe to <%s> nonblocking", a->a_name);
      }
    }

    break;
  }
}

static int idsa_module_log_strexec(char *s)
{
  int i, j, e;
  char **a;

  i = 0;
  j = 1;
  e = 0;
  while (s[i] != '\0') {
    if (isspace(s[i])) {
      j = 1;
    } else {
      if (j) {
	e++;
	j = 0;
      }
    }
    i++;
  }
#ifdef TRACE
  fprintf(stderr, "strexec(): <%s> has %d elements\n", s, e);
#endif
  e++;

  a = malloc(sizeof(char *) * e);
  if (a) {
    i = 0;
    j = 1;
    e = 0;
    while (s[i] != '\0') {
      if (isspace(s[i])) {
	s[i] = '\0';
	j = 1;
      } else {
	if (j) {
	  a[e] = s + i;
	  e++;
	  j = 0;
	}
      }
      i++;
    }
    a[e] = NULL;
    execvp(s, a);
#ifdef TRACE
    fprintf(stderr, "strexec(): execvp(%s,...) failed: %s\n", s, strerror(errno));
#endif
  }
  return -1;
}

/****************************************************************************/

IDSA_MODULE *idsa_module_load_log(IDSA_RULE_CHAIN * c)
{
  IDSA_MODULE *result;

  result = idsa_module_new_version(c, "log", IDSA_MODULE_INTERFACE_VERSION);
  if (result) {
    result->action_start = &idsa_module_log_astart;
    result->action_cache = &idsa_module_log_acache;
    result->action_do = &idsa_module_log_ado;
    result->action_stop = &idsa_module_log_astop;
  }

  return result;
}
